"""
Copyright (c) 2008-2021 synodriver <synodriver@gmail.com>
"""
import os
from random import randint
from unittest import TestCase
import sys
sys.path.append(".")
os.environ["HFM_USE_CFFI"] = "1"

from pyhuffman import decode, encode


class Testall(TestCase):
    def test_encode(self):
        for i in range(1000):
            data = bytes([randint(0, 255) for _ in range(randint(100, 1000))])
            self.assertEqual(decode(encode(data)), data)


if __name__ == "__main__":
    import unittest

    unittest.main()



"""
Copyright (c) 2008-2021 synodriver <synodriver@gmail.com>
"""
import os
from pathlib import Path
from typing import IO, Union

from pyhuffman.backends.cffi._hfm import ffi, lib

InputType = Union[str, bytes, Path, IO]


def ensure_bytes(file) -> bytes:
    if isinstance(file, Path):
        return str(file).encode()
    elif isinstance(file, str):
        return file.encode()
    else:
        return bytes(file)


def encode_file(in_: InputType, out_: InputType) -> int:
    use_fopen = 0
    if hasattr(in_, "fileno") and hasattr(out_, "fileno"):
        in_file = lib.fdopen(os.dup(in_.fileno()), "rb")
        out_file = lib.fdopen(os.dup(out_.fileno()), "wb")
    else:
        use_fopen = 1
        in_name_bytes = ensure_bytes(in_)
        out_name_bytes = ensure_bytes(out_)
        in_file = lib.fopen(ffi.from_buffer(in_name_bytes), "rb")
        if in_file == ffi.NULL:
            raise FileNotFoundError(f"can not open{in_}")

        out_file = lib.fopen(ffi.from_buffer(out_name_bytes), "wb")
        if out_file == ffi.NULL:
            lib.fclose(in_file)
            raise FileNotFoundError(f"can not open{out_}")

    ret: int = lib.huffman_encode_file(in_file, out_file)

    if use_fopen:
        lib.fclose(in_file)
        lib.fclose(out_file)
    if ret > 0:
        raise ValueError
    return ret


def decode_file(in_: InputType, out_: InputType) -> int:
    use_fopen = 0
    if hasattr(in_, "fileno") and hasattr(out_, "fileno"):
        in_file = lib.fdopen(os.dup(in_.fileno()), "rb")
        out_file = lib.fdopen(os.dup(out_.fileno()), "wb")
    else:
        use_fopen = 1
        in_name_bytes = ensure_bytes(in_)
        out_name_bytes = ensure_bytes(out_)
        in_file = lib.fopen(ffi.from_buffer(in_name_bytes), "rb")
        if in_file == ffi.NULL:
            raise FileNotFoundError(f"can not open{in_}")

        out_file = lib.fopen(ffi.from_buffer(out_name_bytes), "wb")
        if out_file == ffi.NULL:
            lib.fclose(in_file)
            raise FileNotFoundError(f"can not open{out_}")

    ret: int = lib.huffman_decode_file(in_file, out_file)

    if use_fopen:
        lib.fclose(in_file)
        lib.fclose(out_file)
    if ret > 0:
        raise ValueError
    return ret


def encode(data: bytes) -> bytes:
    input_size = len(data)
    out_buf = ffi.new("unsigned char **")
    out_len = ffi.new("uint32_t*")

    ret: int = lib.huffman_encode_memory(
        ffi.cast("const unsigned char*", ffi.from_buffer(data)),
        input_size,
        out_buf,
        out_len,
    )
    assert out_buf[0] != ffi.NULL
    out: bytes = ffi.unpack(ffi.cast("char*", out_buf[0]), out_len[0])
    lib.PyMem_Free(out_buf[0])
    if ret > 0:
        raise ValueError
    return out


def decode(data: bytes) -> bytes:
    input_size = len(data)
    out_buf = ffi.new("unsigned char **")
    out_len = ffi.new("uint32_t*")

    ret: int = lib.huffman_decode_memory(
        ffi.cast("const unsigned char*", ffi.from_buffer(data)),
        input_size,
        out_buf,
        out_len,
    )
    assert out_buf[0] != ffi.NULL
    out: bytes = ffi.unpack(ffi.cast("char*", out_buf[0]), out_len[0])
    lib.PyMem_Free(out_buf[0])
    if ret > 0:
        raise ValueError
    return out



import glob
import sys

from cffi import FFI

ffibuilder = FFI()
ffibuilder.cdef(
    """
int huffman_encode_file(FILE* in, FILE* out);

int huffman_decode_file(FILE* in, FILE* out);

int huffman_encode_memory(const unsigned char* bufin,
			  uint32_t bufinlen,
			  unsigned char** pbufout,
			  uint32_t* pbufoutlen);

int huffman_decode_memory(const unsigned char* bufin,
			  uint32_t bufinlen,
			  unsigned char** bufout,
			  uint32_t* pbufoutlen);

FILE *fopen(const char *filename, const char  *opentype);
FILE *fdopen(int fdescriptor, const char *opentype);
int  fclose(FILE *stream);
void* PyMem_Malloc(size_t n);
void PyMem_Free(void *p);
    """
)

source = """
#include <stdint.h>
#include <stdio.h>
#include "huffman.h"
"""

ffibuilder.set_source(
    "pyhuffman.backends.cffi._hfm",
    source,
    sources=[f"./huffman/huffman.c"],
    include_dirs=["./huffman"],
    define_macros=[("PYTHON", None)],
)

if __name__ == "__main__":
    ffibuilder.compile()



# cython: language_level=3
# cython: cdivision=True
# distutils: language=c++
from cpython.bytes cimport PyBytes_FromStringAndSize
from cpython.mem cimport PyMem_Free
from cpython.object cimport PyObject_HasAttrString
from libc.stdint cimport uint8_t, uint32_t
from libc.stdio cimport FILE, fclose, fdopen, fopen

from pyhuffman.backends.cython.huffman cimport (huffman_decode_file,
                                                huffman_decode_memory,
                                                huffman_encode_file,
                                                huffman_encode_memory)

import os
from pathlib import Path


cdef bytes ensure_bytes(object file):
    if isinstance(file, Path):
        return str(file).encode()
    elif isinstance(file, str):
        return file.encode()
    else:
        return bytes(file)

cpdef inline int encode_file(object in_, object out_) except +:
    cdef:
        int use_fopen = 0
        FILE * in_file
        FILE * out_file
    if PyObject_HasAttrString(in_, "fileno") and PyObject_HasAttrString(out_, "fileno"):
        in_file = fdopen(os.dup(in_.fileno()), "rb")
        out_file = fdopen(os.dup(out_.fileno()), "wb")
    else:
        use_fopen = 1
        in_name_bytes = ensure_bytes(in_)
        out_name_bytes = ensure_bytes(out_)
        in_file = fopen(<char*>in_name_bytes, "rb")
        if in_file == NULL:
            raise FileNotFoundError(f"can not open{in_}")

        out_file = fopen(<char *> out_name_bytes, "wb")
        if out_file == NULL:
            fclose(in_file)
            raise FileNotFoundError(f"can not open{out_}")
    cdef int ret
    with nogil:
        ret = huffman_encode_file(in_file, out_file)

    if use_fopen:
        fclose(in_file)
        fclose(out_file)
    if ret > 0:
        raise ValueError
    return ret

cpdef inline int decode_file(object in_, object out_) except +:
    cdef:
        int use_fopen = 0
        FILE * in_file
        FILE * out_file
    if PyObject_HasAttrString(in_, "fileno") and PyObject_HasAttrString(out_, "fileno"):
        in_file = fdopen(os.dup(in_.fileno()), "rb")
        out_file = fdopen(os.dup(out_.fileno()), "wb")
    else:
        use_fopen = 1
        in_name_bytes = ensure_bytes(in_)
        out_name_bytes = ensure_bytes(out_)
        in_file = fopen(<char*>in_name_bytes, "rb")
        if in_file == NULL:
            raise FileNotFoundError(f"can not open{in_}")

        out_file = fopen(<char *> out_name_bytes, "wb")
        if out_file == NULL:
            fclose(in_file)
            raise FileNotFoundError(f"can not open{out_}")
    cdef int ret
    with nogil:
        ret = huffman_decode_file(in_file, out_file)

    if use_fopen:
        fclose(in_file)
        fclose(out_file)
    if ret > 0:
        raise ValueError
    return ret

cpdef inline bytes encode(const uint8_t[::1] data):
    cdef:
        Py_ssize_t input_size = data.shape[0]
        unsigned char * out_buf = NULL
        uint32_t out_len
        int ret
    with nogil:
        ret = huffman_encode_memory(<const unsigned char*>&data[0], <uint32_t>input_size, &out_buf, &out_len)
    assert out_buf != NULL
    cdef bytes out = PyBytes_FromStringAndSize(<char *>out_buf, <Py_ssize_t>out_len)
    PyMem_Free(out_buf)
    if ret > 0:
        raise ValueError
    return out

cpdef inline bytes decode(const uint8_t[::1] data):
    cdef:
        Py_ssize_t input_size = data.shape[0]
        unsigned char * out_buf = NULL
        uint32_t out_len
        int ret
    with nogil:
        ret = huffman_decode_memory(<const unsigned char*>&data[0], <uint32_t>input_size, &out_buf, &out_len)
    assert out_buf != NULL
    cdef bytes out = PyBytes_FromStringAndSize(<char *>out_buf, <Py_ssize_t>out_len)
    PyMem_Free(out_buf)
    if ret > 0:
        raise ValueError
    return out



# cython: language_level=3
# cython: cdivision=True
# distutils: language=c++
from libc.stdint cimport uint32_t
from libc.stdio cimport FILE


cdef extern from "huffman.h" nogil:
    int huffman_encode_file(FILE* in_, FILE* out)

    int huffman_decode_file(FILE* in_, FILE* out)

    int huffman_encode_memory(const unsigned char* bufin,
                  uint32_t bufinlen,
                  unsigned char** pbufout,
                  uint32_t* pbufoutlen)

    int huffman_decode_memory(const unsigned char* bufin,
                  uint32_t bufinlen,
                  unsigned char** bufout,
                  uint32_t* pbufoutlen)


import os
import platform

impl = platform.python_implementation()


def _should_use_cffi() -> bool:
    ev = os.getenv("HFM_USE_CFFI")
    if ev is not None:
        return True
    if impl == "CPython":
        return False
    else:
        return True


if not _should_use_cffi():
    from pyhuffman.backends.cython import *
else:
    from pyhuffman.backends.cffi import *

__version__ = "0.1.3"
